/*
 * Copyright (C) 2009 by Aiwota Programmer
 * aiwotaprog@tetteke.tk
 *
 * This file is part of Dialektos.
 *
 * Dialektos is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Dialektos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Dialektos.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "text_element_plain.hxx"

#include <pangomm/layout.h>
#include <pangomm/attributes.h>
#include <pangomm/attrlist.h>
#include <pangomm/context.h>
#include <pangomm/font.h>
#include <pangomm/rectangle.h>
#include <glibmm/refptr.h>
#include <glibmm/ustring.h>
#include <gdkmm/gc.h>
#include <gdkmm/cursor.h>
#include <gdkmm/rectangle.h>
#include <boost/foreach.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/array.hpp>
#include "text_view_drawing_set.hxx"
#include "text_element_char_size_cache.hxx"


namespace dialektos {

namespace text_element {


/* static data members */
CharSizeCache Plain::char_size_cache;
gdouble Plain::approximate_char_height = -1;


double Plain::get_approximate_char_height(
    Glib::RefPtr<const Pango::Layout> layout) {

  if (approximate_char_height < 0) {
    Glib::RefPtr<Pango::Context> context = layout->get_context();
    Pango::FontDescription desc = context->get_font_description();
    Glib::RefPtr<Pango::Font> font = context->load_font(desc);
    Pango::Rectangle rect = font->get_glyph_logical_extents(0);
    approximate_char_height = double(rect.get_height())/Pango::SCALE*1.2;
  }
  return approximate_char_height;
}

void Plain::trim_right() {
  using namespace boost::lambda;

  boost::trim_right_if(str_, _1 == ' ');
}

void Plain::layout(text_view::LayoutSet& set) {

  /* caching for drawing */
  x_ = set.x;
  y_ = set.y;
  x_start_ = set.x_start;
  x_end_ = set.x_end;

  Pango::AttrList list = Pango::AttrList();
  Pango::AttrInt attr = bold_ ?
      Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD):
      Pango::Attribute::create_attr_weight(Pango::WEIGHT_NORMAL);
  list.insert(attr);
  set.layout->set_attributes(list);

  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size(uch, set.layout, w, h, bold_);

    if (set.x + w > x_end_) {
      // begin new line
      set.x = x_start_;
      set.y += get_approximate_char_height(set.layout);
    }

    set.x += w;
  }
}

void Plain::draw(text_view::DrawingSet& set) const {

  Glib::RefPtr<const Gdk::GC> gc = set.style->get_text_gc(Gtk::STATE_NORMAL);

  Pango::AttrList list = Pango::AttrList();
  Pango::AttrInt attr = bold_ ?
      Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD):
      Pango::Attribute::create_attr_weight(Pango::WEIGHT_NORMAL);
  list.insert(attr);

  set_attributes(list);
  set.layout->set_attributes(list);

  size_t start_index = 0;
  size_t end_index = 0;
  get_selection_start_end_index(set.start_element, set.start_index,
      set.end_element, set.end_index, start_index, end_index);

  gdouble x = x_;
  gdouble y = y_;
  size_t index_count = 0;
  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size_from_cache(uch, w, h, bold_);

    if (x + w > x_end_) {
      // begin new line
      x = x_start_;
      y += approximate_char_height;//(set.layout);
    }

    double deltaY = approximate_char_height - h;
    set.layout->set_text(Glib::ustring(1, uch));

    if (start_index != end_index &&
        index_count >= start_index && index_count < end_index)
      set.window->draw_layout(gc, x, y + deltaY - set.adj_value,
          set.layout,
          set.style->get_fg(Gtk::STATE_SELECTED),
          set.style->get_bg(Gtk::STATE_SELECTED));
    else
      set.window->draw_layout(gc, x, y + deltaY - set.adj_value,
          set.layout);

    x += w;
    ++index_count;
  }
}


void Plain::set_attributes(Pango::AttrList& list) const {
}

bool Plain::is_xy_on_this(const gdouble x, const gdouble y) const {

  gdouble x_pos = x_;
  gdouble y_pos = y_;

  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size_from_cache(uch, w, h, bold_);

    if (x_pos + w > x_end_) {
      // begin new line
      x_pos = x_start_;
      y_pos += approximate_char_height;
    }

    if (x >= x_pos && x < x_pos + w &&
        y >= y_pos && y < y_pos + approximate_char_height) {
      return true;
    }

    x_pos += w;
  }

  return false;
}

bool Plain::is_xy_near_to_this(const gdouble x, const gdouble y) const {
  gdouble x_pos = x_;
  gdouble y_pos = y_;

  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size_from_cache(uch, w, h, bold_);

    if (x_pos + w > x_end_) {
      // begin new line

      if (x >= x_pos && y >= y_pos && y < y_pos + approximate_char_height) {
        // (x, y) is on the spaces caused by wrapping
        return true;
      }

      x_pos = x_start_;
      y_pos += approximate_char_height;
    }

    if (x >= x_pos && x < x_pos + w &&
        y >= y_pos && y < y_pos + approximate_char_height) {
      return true;
    }

    x_pos += w;
  }

  return false;
}


Gdk::Rectangle Plain::xy_to_rectangle(gdouble x, gdouble y) const {
  gdouble x_pos = x_;
  gdouble y_pos = y_;
  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size_from_cache(uch, w, h, bold_);

    if (x_pos + w > x_end_) {
      // begin new line
      x_pos = x_start_;
      y_pos += approximate_char_height;
    }

    if (x >= x_pos && x < x_pos + w &&
        y >= y_pos && y < y_pos + approximate_char_height)
        return Gdk::Rectangle(x_pos, y_pos, w, approximate_char_height);

    x_pos += w;
  }

  return Gdk::Rectangle(x_pos, y_pos, 0, approximate_char_height);
}


int Plain::xy_to_index(const gdouble x, const gdouble y) const {

  gdouble x_pos = x_;
  gdouble y_pos = y_;
  int index = 0;

  BOOST_FOREACH(const gunichar& uch, str_) {
    double w, h;
    char_size_cache.get_char_size_from_cache(uch, w, h, bold_);

    if (x_pos + w > x_end_) {
      // begin new line
      if (x >= x_pos &&
          y >= y_pos && y < y_pos + approximate_char_height) {
        // (x, y) is on the spaces caused by wrapping
        return index;
      }

        x_pos = x_start_;
      y_pos += approximate_char_height;
    }

    if (x >= x_pos && x < x_pos + w &&
        y >= y_pos && y < y_pos + approximate_char_height) {
      if (x <= x_pos + w/2) {
        // left of this character
        return index;
      } else {
        // right of this character
        return index+1;
      }
    }

    x_pos += w;
    ++index;
  }
  return index;
}

Gdk::CursorType Plain::get_cursor_type() const {
  return Gdk::XTERM;
}

void Plain::get_selection_start_end_index(
    const Plain* const start_element, const size_t start_index,
    const Plain* const end_element, const size_t end_index,
    size_t& start, size_t& end) const {

  start = 0;
  end = 0;
  if (start_element && end_element) {
    if (start_element == this && end_element == this) {
      start = start_index;
      end = end_index;
    } else if (start_element == this) {
      start = start_index;
      end = str_.size();
    } else if (end_element == this) {
      end = end_index;
    } else if (start_element->y_ < y_ && end_element->y_ > y_) {
      end = str_.size();
    } else if (start_element->y_ == y_ && end_element->y_ == y_ &&
        start_element->x_ < x_ && end_element->x_ > x_) {
      end = str_.size();
    } else if (start_element->y_ == y_ && end_element->y_ != y_ &&
        start_element->x_ < x_) {
      end = str_.size();
    } else if (end_element->y_ == y_ && start_element->y_ != y_ &&
        end_element->x_ > x_) {
      end = str_.size();
    }
  }

}


} // namespace text_element

} // namespace dialektos
