/*
 * Copyright (C) 2009 by Aiwota Programmer
 * aiwotaprog@tetteke.tk
 *
 * This file is part of Dialektos.
 *
 * Dialektos is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Dialektos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Dialektos.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "thread_window.hxx"

#include <glibmm/convert.h>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/filesystem.hpp>
#include <iostream>
#include <fstream>
#include "bbs_detail_base.hxx"
#include "thread_idx.hxx"
#include "http_get.hxx"
#include "misc.hxx"


namespace dialektos {


void ThreadWindow::create(std::auto_ptr<bbs_detail::Base> bbs) {
  regist(new ThreadWindow(bbs));
}

ThreadWindow::ThreadWindow(std::auto_ptr<bbs_detail::Base> bbs) :
  SuperClass(),
  text_view_(),
  scrolled_(text_view_.get_adjustment()),
  bbs_(bbs), http_getter_(), idx_() {

  set_default_size(400,300);

  scrolled_.add(text_view_);
  add(scrolled_);

  text_view_.signal_button_press_event().connect_notify(
      sigc::mem_fun(*this, &ThreadWindow::on_child_button_press));

  show_all();

  load();
}

bool ThreadWindow::load() {
  using namespace boost::posix_time;
  ptime start = microsec_clock::local_time();

  if (bbs_) {

    bbs_->load_thread(text_view_);
    text_view_.relayout();
    text_view_.queue_draw();

    std::cout <<
    to_simple_string(microsec_clock::local_time() - start) << std::endl;

    idx_ = ThreadIdx::from_xml(bbs_->get_thread_idx_path());
  }

  return false;
}

bool ThreadWindow::is_same(const bbs_detail::Base& bbs) const {
  const bbs_detail::Base& lhs = *bbs_;
  return lhs == bbs;
}

std::string ThreadWindow::get_uri() const {
  return bbs_->get_thread_uri();
}


void ThreadWindow::on_action_view_refresh() {
  if (http_getter_) return;

  statusbar_.push("HTTP/1.0 GET...");

  const std::string uri = bbs_->get_thread_dat_uri();
  http::Header request_header = bbs_->get_thread_dat_request_header();
  if (!request_header.get_range().empty()) {
    request_header.set_if_modified_since(idx_.last_modified_);
    request_header.set_if_none_match(idx_.etag_);
  }

  http_getter_.reset(new http::GetInThread(uri, request_header));
  http_getter_->signal_end().connect(
      sigc::mem_fun(*this, &ThreadWindow::on_http_get_end));
  http_getter_->run();
}

void ThreadWindow::on_action_view_stop() {
  if (http_getter_) http_getter_->cancel();
}

void ThreadWindow::on_http_get_end(bool success) {
//  const std::string uri = http_getter_->get_uri();
//  const http::Header request_header = http_getter_->get_request_header();
  const http::Response response = http_getter_->get_response();
  const boost::system::error_code err = http_getter_->get_error();
  http_getter_.reset(0);
  if (err) {
    statusbar_.push(err.message());
    return;
  }
  if (!success) {
    statusbar_.push("Canceled.");
    return;
  }

  on_refresh_end(response.get_status_line(), response.get_header());

  const int code = response.get_status_line().get_code();
  if (code != 200 && code != 206) return;

  bbs_->load_thread_from_string(response.get_content(), text_view_);
  text_view_.relayout();
  text_view_.queue_draw();

  save_content(response);
}

void ThreadWindow::on_refresh_end(const http::StatusLine& status,
    const http::Header& header) {
  std::string message = status.get_line();
  const std::string last_modified = header.get_last_modified();
  if (!last_modified.empty()) {
    message += " ";
    message += last_modified;
  }
  statusbar_.push(message);
}

void ThreadWindow::save_content(const http::Response& response) {
  const int code = response.get_status_line().get_code();
  if (code != 200 && code != 206) return;

  if (!misc::create_directories(boost::filesystem::path(
      bbs_->get_thread_file_path()).parent_path())) return;
  std::ofstream ofs(bbs_->get_thread_file_path().c_str(),
      code == 200 ?
          std::ios::out | std::ios::trunc : std::ios::out | std::ios::app);
  ofs << response.get_content();
  ofs.close();

  if (code == 200)
    idx_.title_ = bbs_->get_title_from_string(response.get_content());

  idx_.last_modified_ = response.get_header().get_last_modified();
  idx_.etag_ = response.get_header().get_etag();

  idx_.line_count_ = text_view_.get_res_num();

  idx_.to_xml(boost::filesystem::path(bbs_->get_thread_idx_path()));
}


} // namespace dialektos
